import data.setoid.partition

variable {α : Type*}

/-!
# Design note

instead of making `partition` a structure which bundles the collection of sets into it, we instead
create a `Prop`-valued structure `is_partition` which allows us to unbundle statements involving
partitions and simplifies things a bit. Then whenever we want bundled partitions (e.g., when we
establish the equivalence with `setoid`, we just use the subtype
`{S : set (set α) // is_partition S}`).
-/

/-- `is_partition` is the property that a set of sets on a type constitutes a partition.
The sets must be nonempty, they must cover the entire type, and they must be pairwise disjoint. -/
structure is_partition (S : set (set (α))) : Prop :=
(nonempty' : ∀ {s : set α} (hs : s ∈ S), s.nonempty)
(covers' : ∀ x : α, ∃ s, s ∈ S ∧ x ∈ s)
(pairwise_disjoint' : ∀ s₁ s₂ ∈ S, (∃ x, x ∈ s₁ ∧ x ∈ s₂) → s₁ = s₂)

/-- The relation `r : α → α → Prop` generated from a set of sets on `α` where `r x y` if and only
if there is a set in the collection containing both `x` and `y`. -/
def rel_of_sets (S : set (set α)) (x y : α) : Prop :=
∃ s, s ∈ S ∧ x ∈ s ∧ y ∈ s

/-- Given a set of sets on `α` that forms a partition and some `x : α`, choose a set in the
partition containing `x`. This choice will necessarily be unique. -/
def containing_set {S : set (set α)} (h : is_partition S) (x : α) : set α :=
classical.some $ h.covers' x

/-- The `containing_set` is always an element of the partition. -/
lemma containing_set_mem {S : set (set α)} (h : is_partition S) (x : α) :
  containing_set h x ∈ S :=
(classical.some_spec $ h.covers' x).1

/-- An element of `α` is always a member of its containing set. -/
lemma mem_containing_set {S : set (set α)} (h : is_partition S) (x : α) :
  x ∈ containing_set h x :=
(classical.some_spec $ h.covers' x).2

/-- The `containing_set x` is the *unique* element of the partition containing `x`. -/
lemma eq_containing_set_of_mem {S : set (set α)} (h : is_partition S)
  {s : set α} (hs : s ∈ S) {x : α} (hx : x ∈ s) : s = containing_set h x :=
h.pairwise_disjoint' _ hs _ (containing_set_mem h x) ⟨x, hx, mem_containing_set h x⟩

/-- `rel_of_sets` is an equivalence relation when the sets form a partition. -/
lemma equivalence_rel_of_sets {S : set (set α)} (h : is_partition S) :
  equivalence (rel_of_sets S) :=
begin
  refine ⟨λ x, _, λ x y, _, λ x y z hxy hyz, _⟩,
  { let hx := mem_containing_set h x,
    exact ⟨containing_set h x, containing_set_mem h x, hx, hx⟩ },
  { simp only [rel_of_sets, and_comm, imp_self] },
  { obtain ⟨s₁, hs₁, hx, hy₁⟩ := hxy,
    obtain ⟨s₂, hs₂, hy₂, hz⟩ := hyz,
    refine ⟨s₁, hs₁, hx, h.pairwise_disjoint' _ hs₂ _ hs₁ ⟨y, hy₂, hy₁⟩ ▸ hz⟩ }
end

/-- `equiv_class x` is the collection of things related to `x` via `r`. -/
def equiv_class (r : α → α → Prop) (x : α) := {y | r x y}

/-- Given a relation `r : α → α → Prop`, produce a set of sets of `α` called `sets_of_rel`
which consists of all the `equiv_class`es. -/
def sets_of_rel (r : α → α → Prop) : set (set α) :=
set.range (equiv_class r)

/-- If `r` is a reflexive relation, every element is a member of its own equivalence class. -/
lemma self_mem_equiv_class {r : α → α → Prop} (hr : reflexive r) (x : α) :
  x ∈ equiv_class r x :=
hr x

/-- If `r` is a symmetric relation and `x` is in the equivalence class of `x`, then `x` is in the
equivalence class of `y`. -/
lemma mem_equiv_class_of_mem_equiv_class {r : α → α → Prop} (hr : symmetric r) (x y : α)
  (hx : x ∈ equiv_class r y) : y ∈ equiv_class r x :=
hr hx

/-- If `r` is a transitive relation and `r x y`, then the equivalence class of `x` is contained in
equivalence class of `y`. -/
lemma subset_equiv_class_of_mem {r : α → α → Prop} (hr : transitive r) (x y : α)
  (hx : x ∈ equiv_class r y) : equiv_class r x ⊆ equiv_class r y :=
λ z hz, hr hx hz

/-- If `r` is a symmetric and transitive relation and `r y x`, then the equivalence classes of `x`
and `y` coincide. -/
lemma eq_equiv_class_of_mem {r : α → α → Prop} (hr : symmetric r) (hr' : transitive r) (x y : α)
  (hx : x ∈ equiv_class r y) : equiv_class r x = equiv_class r y :=
le_antisymm (λ z hz, hr' hx hz) (λ z hz, hr' (hr hx) hz)

/-- If `r` is an equivalence relation then the equivalence classes `sets_of_rel` form a partition.-/
lemma is_partition_sets_of_rel {r : α → α → Prop} (hr : equivalence r) :
  is_partition (sets_of_rel r) :=
{ nonempty' :=
  begin
    rintro - ⟨x, rfl⟩,
    exact ⟨x, self_mem_equiv_class hr.1 x⟩,
  end,
  covers' := λ x, ⟨equiv_class r x, ⟨x, rfl⟩, self_mem_equiv_class hr.1 x⟩,
  pairwise_disjoint' :=
  begin
    rintro - ⟨x, rfl⟩ - ⟨y, rfl⟩ ⟨z, hzx, hzy⟩,
    exact (eq_equiv_class_of_mem hr.2.1 hr.2.2 z x hzx).symm.trans
      (eq_equiv_class_of_mem hr.2.1 hr.2.2 z y hzy),
  end }

/-- `partition` is the `equiv` (denoted `≃`) between the type `setoid α` of bundled equivalence
relations and the subtype `{S : set (set α) // is_partition S}` of bundled partitions. The function
and its inverse are given by (bundled versions of) `sets_of_rel` and `rel_of_sets`.

We give it this name so that if `s : setoid α` then `partition s` is the partition generated by
the equivalence relation. We will use this feature later. -/
@[simps]
def partition : setoid α ≃ {S : set (set α) // is_partition S} :=
{ to_fun := λ R, subtype.mk (sets_of_rel R.r) (is_partition_sets_of_rel R.iseqv),
  inv_fun := λ S, setoid.mk (rel_of_sets S) (equivalence_rel_of_sets S.prop),
  right_inv :=
  begin
    refine λ S, subtype.ext (set.ext $ λ s, _),
    simp only [subtype.coe_mk, sets_of_rel],
    split,
    { rintro ⟨x, rfl⟩,
      obtain ⟨t, ht, hxt⟩ := S.prop.covers' x,
      convert ht,
      ext y,
      refine ⟨_, λ hy, ⟨t, ht, hxt, hy⟩⟩,
      rintro ⟨s, hs, hxs, hys⟩,
      exact S.prop.pairwise_disjoint' _ hs _ ht ⟨x, hxs, hxt⟩ ▸ hys },
    { intro hs,
      obtain ⟨x, hxs⟩ := S.prop.nonempty' hs,
      refine ⟨x, _⟩,
      ext y,
      refine ⟨_, λ hy, ⟨s, hs, hxs, hy⟩⟩,
      rintro ⟨t, ht, hxt, hy⟩,
      exact S.prop.pairwise_disjoint' _ ht _ hs ⟨x, hxt, hxs⟩ ▸ hy },
  end,
  left_inv :=
  begin
    refine λ R, setoid.ext (λ x y, _),
    simp only [rel_of_sets, subtype.coe_mk, sets_of_rel],
    refine ⟨_, λ hxy, ⟨_, ⟨x, rfl⟩, R.iseqv.1 x, hxy⟩⟩,
    rintro ⟨s, ⟨z, rfl⟩, hx, hy⟩,
    exact R.iseqv.2.2 (R.iseqv.2.1 hx) hy,
  end }

.

variable {β : Type*}

/-!
# Quotient types

Lean has built-in support for quotient types. If `s : setoid α`, then `quotient s` should be
considered as the type of *abstract equivalence classes* under the equivalence relation. These
behave exactly like equivalence classes (e.g., for `x y : α`, `x ≈ y ↔ ⟦x⟧ = ⟦y⟧`), except
`⟦x⟧ ⟦y⟧ : quotient s`; they do *not* have type `set α`. This is why we call them *abstract*.

That being said, below we show that there is a natural `equiv` between the types `quotient s`
and `partition s`, where the latter is the type of *concrete equivalence classes* of `s`.
Technically speaking, a term `t : partition s` doesn't have type `set α`, however, it does have
a *coercion* to `set α`, so it is reasonable to consider it as a set. In contrast, there is no
such coercion to `set α` for `q : quotient s`.

There are several key declarations related to `quotient` types, and these are:

#check @quotient.sound
#check @quotient.exact
#check @quotient.induction_on
#check @quotient.lift
#check @quotient.lift_mk
#check @quotient.map

Note that `quotient.lift` and `quotient.lift_eq` (the latter appears below) together make up the
universal property of the quotient. This is made explicit in `quotient_is_quotient`.
-/

/-- `quotient.lift f hf` is the *unique* function from `quotient s` which makes the diagram commute:

   x ↦ ⟦x⟧
α ----------> quotient s
  \_          |
    \_        | quotient.lift f hf
      \_      |
   f    \_    V
          \-> β
-/
lemma quotient.lift_eq [s : setoid α] (f : α → β) (hf : ∀ x y, x ≈ y → f x = f y)
  (g : quotient s → β) (hg : (λ x, g ⟦x⟧) = f) : g = quotient.lift f hf :=
funext $ λ q, quotient.induction_on q (λ x, (congr_fun hg x).trans (quotient.lift_mk f hf x))

/-- Given a bundled equivalence relation (an `s : setoid α`), `to_equiv_class` is the the map
which sends `x : α` to the set `t : partition s` which is the equivalence class of `x`.

Here we use the "coinductive" property of `subtype`, namely `subtype.coind` which allows us to
construct functions *into* a subtype. -/
def to_equiv_class [s : setoid α] : α → partition s :=
subtype.coind (equiv_class s.r) (λ x, ⟨x, rfl⟩)

/-- `quotient_to_partition` is a function which takes an element of the quotient and maps it to
an element of the corresponding partition. Effectively, this takes an abstract equivalence class
`⟦x⟧` to a concrete equivalence class `@to_equiv_class α s x`.

This uses `quotient.lift` to construct a function *out of* `quotient s`. -/
def quotient_to_partition [s : setoid α] : quotient s → partition s :=
quotient.lift (@to_equiv_class α s) $
  λ x y hxy, subtype.ext $ eq_equiv_class_of_mem s.iseqv.2.1 s.iseqv.2.2 x y (s.iseqv.2.1 hxy)

/-- `rep` takes a concrete equivalence class `t : partition s` and chooses an arbitrary
representative from it. -/
noncomputable def rep [s : setoid α] (t : partition s) : α :=
classical.some ((partition s).prop.nonempty' t.prop)

/-- The representative `rep t` is always an element of `t` (where we view `t` as a set). -/
lemma rep_mem [s : setoid α] (t : partition s) : rep t ∈ (t : set α) :=
classical.some_spec ((partition s).prop.nonempty' t.prop)

/-- The concrete equivalence class of `rep t` is `t` itself. -/
lemma equiv_class_rep [s : setoid α] (t : partition s) :
  equiv_class s.r (rep t) = t :=
(partition s).prop.pairwise_disjoint' _ ⟨rep t, rfl⟩ _ t.prop
  ⟨rep t, self_mem_equiv_class s.iseqv.1 (rep t), rep_mem t⟩

/-- The natural equivalence between abstract equivalence classes and conrete equivalence classes. -/
noncomputable def equiv.quotient_partition [s : setoid α] :
  quotient s ≃ partition s :=
{ to_fun := quotient_to_partition,
  inv_fun := λ t, ⟦rep t⟧,
  left_inv := @quotient.ind α s (λ q, ⟦rep (quotient_to_partition q)⟧ = q) $
    λ x, quotient.sound (s.iseqv.2.1 (rep_mem (quotient_to_partition ⟦x⟧))),
  right_inv := λ t, subtype.ext (equiv_class_rep t) }

/-!
# Abstract quotients and the universal property

In this section we define a class `is_quotient q` where `q : α → β` is a map and `α` is equipped
with a bundled equivalence relation `s : setoid α`. `is_quotient q` means that `β` should be
thought of as the type of `α` where we quotient out by the equivalence relation `s`, i.e., where
we identify elements that are related.

One should think of `is_quotient` as an even further abstraction of the idea of quotient types.
Lean doesn't implement quotient types in this way primarily for performance and computability
reasons. -/

variables [s : setoid α] {q : α → β}
include s

/-- `is_quotient q` says two things:
1. `q` idenitifies points related by the equivalence relation `s`.
2. The *universal property of the quotient*: this gives a way to construct *unique* functions out of
the quotient type `β` into an arbitrary type `γ` by providing a function `f : α → γ` defined on the
prequotient `α` satisfying a compatibility condition.

Diagrammatically, a type `β` is an abstract quotient of `α` with respect to the map `q : α → β` if
this diagram commutes whenever `f` satisfies the compatibility condition:

    q
α -----> β
  \_     | ∃! g
    \_   V
   f  \> γ
-/
class is_quotient (q : α → β) : Prop :=
(sound : ∀ x y, x ≈ y → q x = q y)
(exists_unique : ∀ ⦃γ : Type*⦄ (f : α → γ) (hf : ∀ x y, x ≈ y → f x = f y), ∃! g, g ∘ q = f)

/-- Procure the function `g : β → γ` obtained by applying the universal property of the quotient.-/
noncomputable
def is_quotient.fun (h : is_quotient q) {γ : Type*} (f : α → γ) (hf : ∀ x y, x ≈ y → f x = f y) :
  β → γ :=
classical.some $ h.exists_unique f hf

/-- The function obtained from the universal property of the quotient composed with the quotient
map `q` is equal to the function `f : α → γ` on the prequotient from which it is built. -/
lemma is_quotient.fun_comp (h : is_quotient q) {γ : Type*} (f : α → γ)
  (hf : ∀ x y, x ≈ y → f x = f y) : h.fun f hf ∘ q = f :=
(classical.some_spec $ h.exists_unique f hf).1

/-- The function obtained from the universal property of the quotient is the unique function
satisfying the property `is_quotient.fun_comp`, i.e., which makes the diagram commute. -/
lemma is_quotient.fun_unique (h : is_quotient q) {γ : Type*} (f : α → γ)
  (hf : ∀ x y, x ≈ y → f x = f y) :
  ∀ g : β → γ, g ∘ q = f → g = h.fun f hf :=
(classical.some_spec $ h.exists_unique f hf).2

/-- If `β` and `γ` and abstract quotients of `α` with quotient maps `q` and `q'`, then there is a
natrual equivalence between `β` and `γ`. In fact, `is_quotient.equiv` is the *unique* bijection
between these two types which factors through both quotient maps `q` and `q'`. Consequently,
all abstract quotients of `α` by an equivalence relation `s : setoid α` are essentially unique. -/
noncomputable
def is_quotient.equiv {α β γ : Type*} [s : setoid α] {q : α → β} (h : is_quotient q)
  {q' : α → γ} (h' : is_quotient q') : β ≃ γ :=
{ to_fun := h.fun q' h'.sound,
  inv_fun := h'.fun q h.sound,
  left_inv := λ x,
  begin
    have key : h'.fun q h.sound ∘ h.fun q' h'.sound ∘ q = q,
      by rw [h.fun_comp q' h'.sound, h'.fun_comp q h.sound],
    have h₁ := h.fun_unique q h.sound (h'.fun q h.sound ∘ h.fun q' h'.sound) key,
    have h₂ := h.fun_unique q h.sound id (funext (λ x, rfl)),
    exact congr_fun (h₁.trans h₂.symm) x,
  end,
  right_inv := λ x,
  begin
    have key : h.fun q' h'.sound ∘ h'.fun q h.sound ∘ q' = q',
      by rw [h'.fun_comp q h.sound, h.fun_comp q' h'.sound],
    have h₁ := h'.fun_unique q' h'.sound (h.fun q' h'.sound ∘ h'.fun q h.sound) key,
    have h₂ := h'.fun_unique q' h'.sound id (funext (λ x, rfl)),
    exact congr_fun (h₁.trans h₂.symm) x,
  end }

/-- `is_quotient.equiv` factors through the quotient map `q`. -/
lemma is_quotient.equiv_comp {α β γ : Type*} [s : setoid α] {q : α → β} (h : is_quotient q)
  {q' : α → γ} (h' : is_quotient q') : (h.equiv h') ∘ q = q' :=
h.fun_comp q' h'.sound

/-- `(is_quotient.equiv _ _).symm` factors through the quotient map `q'`. -/
lemma is_quotient.equiv_symm_comp {α β γ : Type*} [s : setoid α] {q : α → β} (h : is_quotient q)
  {q' : α → γ} (h' : is_quotient q') : (h.equiv h').symm ∘ q' = q :=
h'.fun_comp q h.sound

/-- The equivalence `is_quotient.equiv` between two abstract quotients is the *unique* equivalence
that factors through the quotient map `q`. -/
noncomputable instance is_quotient.equiv_unique {α β γ : Type*} [s : setoid α] {q : α → β}
  [h : is_quotient q] {q' : α → γ} [h' : is_quotient q'] : unique ({e : β ≃ γ // e ∘ q = q'}) :=
{ default := ⟨h.equiv h', h.equiv_comp h'⟩,
  uniq := λ e, subtype.ext (equiv.ext (λ x, congr_fun (h.fun_unique _ _ _ e.prop) x)) }

/-- If `s : setoid α` is a bundled equivalence relation, then `quotient s` with the map
`(x : α) ↦ (⟦x⟧ : quotient s)` is an abstract quotient of `α`. -/
instance quotient_is_quotient (s : setoid α) :
  is_quotient (λ x, (⟦x⟧ : quotient s)) :=
{ sound := λ _ _, quotient.sound,
  exists_unique := λ γ f hf, ⟨quotient.lift f hf, by simp, λ g hg, quotient.lift_eq f hf g hg⟩ }

/-- If `s : setoid α` is a bundled equivalence relation, then `partition s` with the map
`(x : α) ↦ (to_equiv_class x : partition s)` is an abstract quotient of `α`. -/
instance partition_is_quotient (s : setoid α) :
  is_quotient (to_equiv_class : α → partition s) :=
{ sound := λ x y hxy, subtype.ext (eq_equiv_class_of_mem s.iseqv.2.1 s.iseqv.2.2 x y $ s.iseqv.2.1 hxy),
  exists_unique :=
  begin
    refine λ γ f hf, ⟨f ∘ rep, funext (λ x, _), _⟩,
    { simpa only [function.comp_app] using (hf _ _ (rep_mem (to_equiv_class x))).symm },
    { intros g hg,
      ext t,
      have h_rep : to_equiv_class (rep t) = t := subtype.ext (equiv_class_rep t),
      simpa only [function.comp_app, h_rep] using congr_fun hg (rep t) }
  end }

/-- Using the bijection between abstract quotients, we can build a new equivalence between
`quotient s` and `partition s` instead of using `equiv.quotient_partition`. -/
noncomputable example : quotient s ≃ partition s :=
(quotient_is_quotient s).equiv (partition_is_quotient s)

/-- The old bijection between `quotient s` and `partition s` is equal to the new one. -/
example : equiv.quotient_partition = (quotient_is_quotient s).equiv (partition_is_quotient s) :=
-- `equiv.quotient_partition` factors through the quotient map `x ↦ ⟦x⟧`, and so they must be equal
let this : {e : quotient s ≃ partition s // e ∘ (λ x, ⟦x⟧) = to_equiv_class} :=
  ⟨equiv.quotient_partition, rfl⟩ in
  congr_arg coe (subsingleton.elim this default)
